@@ -1,13 +1,14 @@
 #!/usr/bin/perl
-# DocumentId:	$Id: debarchiver.pl 2372 2006-12-08 21:04:56Z ola $
+# DocumentId:	$Id: debarchiver.pl 2505 2007-10-09 20:49:22Z ola $
 # Author:	$Author: ola $
-# Date:		$Date: 2006-12-08 22:04:56 +0100 (fre, 08 dec 2006) $
+# Date:		$Date: 2007-10-09 22:49:22 +0200 (tis, 09 okt 2007) $
 # Version:	$Revision: 1.49 $
 # Summary:
 #	This program reads a config file, traverse through all .changes files
 #	in the specified directory and puts the files into the right place.
 #
-#    Copyright (C) 2000-2006 Ola Lundqvist <opal@debian.org>
+#    Copyright (C) 2000-2007 Ola Lundqvist <opal@debian.org>
+#    Copyright (C) 2007      Turbo Fredriksson <turbo@bayour.com>
 #    Copyright (C) 2004-2005 Russ Allbery <rra@stanford.edu>
 #    Copyright (C) 2005      Joel Baker <lucifer@lightbearer.com>
 #    Copyright (C) 2005      Daniel Leidert <daniel.leidert@wgdd.de>
@@ -38,6 +39,7 @@
 ############################# USES ############################################
 ###############################################################################
 
+use Digest::MD5;
 use File::Path qw(mkpath);
 use OpaL::action qw(pdebug action cmdaction
 		    setDebugLevel
@@ -68,6 +70,11 @@ use OpaL::read qw(readfile readcommand);
 #	Fix signcmd default.
 #  2005-10-14 Daniel Leidert <daniel.leidert@wgdd.de>
 #	Remove signcmd variable.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Allow forcing an install of an upload (i.e., ignore existing files in
+#       destdir).
+#  2007-10-09 Ola Lundqvist <opal@debian.org>
+#       Changed force option to ignoredestcheck option.
 
 $copycmd = "cp -af";
 $rmcmd = "rm -f";
@@ -93,6 +100,10 @@ $gpgpassfile = "$ENV{HOME}/.gnupg/passphrase";
 # Changelog:
 #  2000-2004 Ola Lundqvist <opal@debian.org>
 #	Written
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Updated with new options.
+#  2007-10-09 Ola Lundqvist <opal@debian.org>
+#       Changed force option to ignoredestcheck option.
 
 # Indexed by config name.
 %CConf = ();
@@ -136,11 +147,15 @@ $majordefault = 'main';
 
 @mailtos = ();
 
+$mailfrom = "";
+
 @ARGS1 = @ARGV;
 @ARGS2 = ();
 
+$incompletetime = 24*60*60;
 $sortpackages = 1;
 $fixstructure = 1;
+$ignoredestcheck = 0;
 
 ###############################################################################
 ############################# ARGS ############################################
@@ -199,25 +214,28 @@ while ($_ = shift @ARGS1) {
 #  2005-10-08 Ola Lundqvist <opal@debian.org>
 #	Added the possibility to specify a config file on command line.
 #	Inspired by a patch from Alexander Wirt <formorer@debian.org>.
+#  2007-08-14 Ola Lundqvist <opal@debian.org>
+#       Changed severity level for configuration file problems from info
+#       to warning, so that it is displayed in the normal case.
 
 if (-e $etcconfigfile) {
     my $t = do $etcconfigfile;
     unless ($t) {
-	pdebug(4, "Loading config file $etcconfigfile:\n\t$!\n\t$@");
+	pdebug(3, "Loading config file $etcconfigfile:\n\t$!\n\t$@");
     }
 }
 
 if (-e $userconfigfile) {
     my $t = do $userconfigfile;
     unless ($t) {
-	pdebug(4, "Loading config file $userconfigfile:\n\t$!\n\t$@");
+	pdebug(3, "Loading config file $userconfigfile:\n\t$!\n\t$@");
     }
 }
 
 if (-e $cmdlineconfigfile) {
     my $t = do $cmdlineconfigfile;
     unless ($t) {
-	pdebug(4, "Loading config file $cmdlineconfigfile:\n\t$!\n\t$@");
+	pdebug(3, "Loading config file $cmdlineconfigfile:\n\t$!\n\t$@");
     }
 }
 
@@ -227,13 +245,16 @@ if (-e $cmdlineconfigfile) {
 # Changelog:
 #  2000-2004 Ola Lundqvist <opal@debian.org>
 #	Written
+#  2007-08-14 Ola Lundqvist <opal@debian.org>
+#       Changed severity level for configuration file problems from info
+#       to warning, so that it is displayed in the normal case.
 
 action(! chdir $inputdir, "Change to dir $inputdir", 2);
 
 if (-e $inputconfigfile) {
     my $t = do $inputconfigfile;
     unless ($t) {
-	pdebug(4, "Loading config file $inputconfigfile:\n\t$!\n\t$@");
+	pdebug(3, "Loading config file $inputconfigfile:\n\t$!\n\t$@");
     }
 }
 
@@ -272,8 +293,15 @@ else {
 #       Default major section function.
 #  2006-11-28 Ola Lundqvist <opal@debian.org>
 #       Sorted options in alphabetical order.
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Make it possible to specify mail sender.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Allow forcing an install of an upload (i.e., ignore existing files in
+#       destdir).
+#  2007-10-09 Ola Lundqvist <opal@debian.org>
+#       Changed force option to ignoredestcheck option.
 
-$version = "0.7.0";
+$version = "0.8.0";
 $versionstring = "Debian package archiver, version $version";
 
 $help =
@@ -322,7 +350,10 @@ options:
                          are installed.  Use this *or* --autoscan, not both.
  --instcmd		DEPRICATED!
  --lockfile file	The lockfile to use, default $lockfile.
+ --incompletetime       The time to allow .changes file to be incomplete in
+                         seconds. Defaults to 24 hours.
  --mailcmd		The mail command to use, default $mail.
+ --mailfrom             Specify mail sender.
  --majordefault section Default major section to use if undefined
                         (default: main)
  --movecmd		Command to move files (currently not used at all).
@@ -340,6 +371,8 @@ options:
                          you know what you are doing.
  --scanonly             Same as --nosort --nostructurefix.
  -v | --version		Prints the version string.
+ --ignoredestcheck	Force install of changes files even if some files
+                        already exists with wrong size or md5 hash.
 
 You can also place config files with the following names (in following order)
 $etcconfigfile, $userconfigfile and $inputconfigfile
@@ -361,9 +394,15 @@ will be parsed. Here you can change the following variables
         \$gpgpassfile   File to provide password to GnuPG.
 	\$inputdir	The input directory (no effect in $inputconfigfile).
 	\$lockfile	The lockfile to use, default $lockfile.
+        \$ignoredestcheck Force install of changes files even if some files
+                         already exists with wrong size or md5 hash.
+                         Default to 0.
+        \$incompletetime The time to allow .changes file to be incomplete in
+                         seconds. Defaults to 24 hours.
 	\@mailtos	The fields in .changes file that should be used for
 			 mailing SUCCESS and REJECT messages. If there is an
 			 @ char in the arrach it will be used directly.
+        \$mailfrom      Specify mail sender.
 	\$majordefault  Default major section (see --majordefault above).
 	\$movecmd	The move command (see --movecmd).
         \%release	Additional information to add to generated Release
@@ -385,6 +424,11 @@ will be parsed. Here you can change the following variables
 #  2006-02-25 Yaroslav Halchenko <debian@onerussian.com> and
 #             Ola Lundqvist <opal@debian.org>
 #       Default major section function.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Allow forcing an install of an upload (i.e., ignore existing files in
+#       destdir).
+#  2007-10-09 Ola Lundqvist <opal@debian.org>
+#       Changed force option to ignoredestcheck option.
 
 while ($_ = shift @ARGS2) {
     if (/^-[a-zA-Z0-9]/) {
@@ -419,6 +463,12 @@ while ($_ = shift @ARGS2) {
 	elsif (/^--mailcmd$/) {
 	    $mailcmd = shift @ARGS2;
 	}
+	elsif (/^--mailfrom$/) {
+	    $mailfrom = shift @ARGS2;
+	}
+	elsif (/^--incompletetime$/) {
+	    $incompletetime = shift @ARGS2;
+	}
 	elsif (/^--rmcmd$/) {
 	    $rmcmd = shift @ARGS2;
 	}
@@ -493,6 +543,9 @@ while ($_ = shift @ARGS2) {
 	    undef $sortpackages;
 	    undef $fixstructure;
 	}
+	elsif (/^--ignoredestcheck$/) {
+	    $ignoredestcheck = 1;
+	}
 	elsif (/^--nosort$/) {
 	    undef $sortpackages;
 	}
@@ -648,13 +701,14 @@ sub createRelease($$$$) {
 #  2003-06-10 Ola Lundqvist <opal@debian.org>
 #	Switched from direct determination of changes owner to use
 #	precalculated data from CMeta{FileOwner}.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub determineMailTo() {
     my $to = "";
-    my $toi;
     my %to_hash;
-    foreach $toi (@mailtos) {
+    foreach my $toi (@mailtos) {
 	# Expand to email if there is no email address.
 	if ($toi !~ /\@/) {
 	    $toi = $CConf{$toi};
@@ -689,21 +743,36 @@ sub determineMailTo() {
 #	Added debugging information and fixed arg handling.
 #  2003-03-13 Ola Lundqvist <opal@debian.org>
 #	Debugged the mail command. It stalls.
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Make it possible to specify mail sender.
 ###############################################################################
 
 sub email($$$$) {
     my ($to, $package, $key, $message) = @_;
     if (length($to) > 0) {
 	pdebug(5, "Executing mail command, $mailcmd -s '$package $key' $to.");
-	if (open(M, "|$mailcmd -s '$package $key' '$to'")) {
-	    print M $message;
-	    close(M);
+	if ($mailfrom ne "") {
+	    if (open(M, "|$mailcmd -s '$package $key' '$to' -- -f '$mailfrom'")) {
+		print M $message;
+		close(M);
+	    }
+	    else {
+		pdebug(2,
+		       "Error executing mail command, $mailcmd -s '$package $key' '$to' -- -f '$mailfrom'.");
+	    }
+	    pdebug(5, "Mail exec done.");
 	}
 	else {
-	    pdebug(2,
-		   "Error executing mail command, $mailcmd -s '$package $key' $to.");
+	    if (open(M, "|$mailcmd -s '$package $key' '$to'")) {
+		print M $message;
+		close(M);
+	    }
+	    else {
+		pdebug(2,
+		       "Error executing mail command, $mailcmd -s '$package $key' '$to'.");
+	    }
+	    pdebug(5, "Mail exec done.");
 	}
-	pdebug(5, "Mail exec done.");
     }
     else {
 	pdebug(3, "No one to send mail to.");
@@ -741,16 +810,20 @@ sub mailSuccess() {
 #	Written using data from mailSuccess.
 #  2003-02-12	Ola Lundqvist <opal@debian.org>
 #	Uses CConf instead of cfile argument.
-#  2003-03-14 Ola Lundqvist <opal@debian.org>
+#  2003-03-14   Ola Lundqvist <opal@debian.org>
 #	Switched to using CMeta for ChangeLog meta information.
+#  2007-10-08   Turbo Fredriksson <turbo@bayour.com>
+#       Error string in beginning and full message (+= => .=).
 ###############################################################################
 
 sub mailReject() {
     # OOPS! We can not read that file after it has been moved!
-    my $message = $CMeta{ChangesContent};
+    my $message;
     if (length($CConf{ERROR}) > 0) {
-	$message += "\n$CConf{ERROR}";
+	$message = "ERROR:\n$CConf{ERROR}\n";
     }
+    $message .= $CMeta{ChangesContent};
+
     pdebug(5, "Mail Reject.");
     email(determineMailTo(),
 	  $CConf{'Source'},
@@ -770,14 +843,15 @@ sub mailReject() {
 # Changelog:
 #  2001-07-23 Ola Lundqvist <opal@debian.org>
 #	Written.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub handleScanAll() {
-    my ($d, $s, $a);
-    foreach $d (@distributions) {
-	foreach $s (@sections) {
+    foreach my $d (@distributions) {
+	foreach my $s (@sections) {
 	    if (-e "$destdir/$d/$s/override") {
-		foreach $a (@architectures) {
+		foreach my $a (@architectures) {
 		    $dests{"$d/$s/binary-$a"} = 1;
 		}
 		$dests{"$d/$s/binary-all"} = 1;
@@ -913,14 +987,16 @@ sub handleScan() {
 # Changelog:
 #  2004-08-10 Ola Lundqvist <opal@debian.org>
 #      Written.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub findSectionsArchitectures($\@\@) {
     my ($dir, $sectionlist, $archlist) = @_;
     my (%dirsections, %dirarches);
-    foreach $s (@sections) {
+    foreach my $s (@sections) {
 	if (-e "$dir/$s/override") {
-	    foreach $a (@architectures) {
+	    foreach my $a (@architectures) {
 		$dirarches{$a} = 1;
 	    }
 	    $dirarches{all} = 1;
@@ -950,6 +1026,8 @@ sub findSectionsArchitectures($\@\@) {
 #	Modified to add bzip2 support.
 #  2005-05-02 Ola Lundqvist <opal@debian.org>
 #	Bugfix for bzip2 support.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub generateIndexConfig() {
@@ -988,7 +1066,7 @@ sub generateIndexConfig() {
     # distributions that are symlinks so that we don't index the same
     # distribution more than once.
     my %changedist = map { s%/.*%%; $_ => 1 } keys %dests;
-    foreach $d (keys %changedist) {
+    foreach my $d (keys %changedist) {
 	my $codename = $distmapping{$d} || $d;
 	next if -l "$destdir/$codename";
 	print CONF "Tree \"dists/$d\" {\n";
@@ -1101,6 +1179,8 @@ sub generateRelease($$\@\@) {
 #	Bugfix for bzip2 support.
 #  2005-08-20 Russ Allbery <rra@stanford.edu>
 #       Call generateRelease for Release file support (optionally signed).
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub handleIndex() {
@@ -1125,11 +1205,11 @@ sub handleIndex() {
     # section, so as to not append to Packages more than once even if we revist
     # the same area twice due to a symlink.
     my %changedist = map { s%/.*%%; $_ => 1 } keys %dests;
-    foreach $d (keys %changedist) {
+    foreach my $d (keys %changedist) {
 	my (@dsections, @darches);
 	&findSectionsArchitectures("$destdir/$d", \@dsections, \@darches);
 	if (-s "$d/Contents-all") {
-	    foreach $a (@darches) {
+	    foreach my $a (@darches) {
 		next if ($a eq 'source' || $a eq 'all');
 		action(! open(ARCH, ">> $d/Contents-$a"),
 		       "Append to $d/Contents-$a",
@@ -1146,7 +1226,7 @@ sub handleIndex() {
 	    }
 	}
 	unlink("$d/Contents-all", "$d/Contents-all.gz");
-	foreach $s (@dsections) {
+	foreach my $s (@dsections) {
 	    if (-s "$d/$s/binary-all/Packages") {
 		foreach $_ (@darches) {
 		    $a = $_;
@@ -1257,6 +1337,8 @@ sub findAndSortChangesFiles($;$) {
 #  2005-05-06 Daniel Leidert <daniel.leidert@wgdd.de>
 #       Add signatur checking setting to arguments given to
 #       findAndSortChangesFiles().
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 # Uses:		A lot.
 ###############################################################################
 
@@ -1305,6 +1387,8 @@ sub handleSorting() {
 #	Moved from START section to this subprocedure.
 #  2004-08-09 Ola Lundqvist <opal@debian.org>
 #	Call createRelease to build Release files.
+#  2007-10-08 Turbo Fredriksson <turbo@bayour.com>
+#       Distmapping for files.
 ###############################################################################
 
 sub handleStructureFix() {
@@ -1313,7 +1397,7 @@ sub handleStructureFix() {
     if (defined $fixstructure) {
 	my ($di, $se, $ar);
 	for $di (@distributions) {
-	    my $dis = $distmapping{$di};
+	    my $dis = $distmapping{$di} || $di;
 	    if (! defined($dis)) {
 		$dis = $di;
 	    }
@@ -1362,18 +1446,99 @@ sub handleStructureFix() {
 # Changelog:
 #  2001-06-29 Ola Lundqvist <opal@debian.org>
 #	Written.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub handleDebOnlyFile($$) {
     my ($kfile, $distr) = @_;
 
     parseDebOnlyFile($kfile, $distr);
-    my $file;
-    foreach $file (keys %CFiles) {
+    foreach my $file (keys %CFiles) {
 	handlePackageFile($file);
     }
 }
 
+###############################################################################
+# Name:		checkFileInDestDirs
+# Description:  Check the location of the file in the dest dirs.
+###############################################################################
+# Uses: parseFileName
+#       parseFileStruct
+#	parseSection
+#       %CConf
+#       %CFiles
+#       %distmapping
+# Arguments:
+#	filename to check for
+# Returns:
+#	filepath     If file exist in dest dirs.
+#       ""           If file not found.
+# Changelog:
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Write function based on the code from Turbo Fredriksson <turbo@bayour.com>
+#       in function verifyChangesFile below.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
+###############################################################################
+
+sub checkFileInDestDirs($) {
+    my ($file) = @_;
+    my (undef, undef, $section, undef) = parseFileStruct($CFiles{$file});
+    my ($major, $section) = parseSection($section);
+
+    my $distrd = $CConf{Distribution};
+    my $inst_files = "";
+
+    foreach my $distr (split /\s+/, $distrd) {
+	$distr = $distmapping{$distr} || $distr;
+	my (undef, undef, $arch, undef) = parseFileName($file);
+	
+	my $archsec = "source";
+	if ($arch !~ /^source$/) {
+	    $archsec = "binary-$arch";
+	}
+	
+	my $inst_file = "$destdir/$distr/$major/$archsec/$section/$file";
+	if (-f "$inst_file") {
+	    if ($inst_files eq "") {
+		$inst_files = $inst_file;
+	    }
+	    else {
+		$inst_files .= " $inst_file";
+	    }
+	}
+    }
+    return $inst_files;
+}
+
+###############################################################################
+# Name:		getMD5HashForFile	
+# Description:  Get the MD5 hash for a file. File must exist.
+###############################################################################
+# Uses:
+#	Digest::MD5
+# Arguments:
+#	filename to check for
+# Returns:
+#	MD5 sum for file
+# Changelog:
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Write function based on the code from Turbo Fredriksson <turbo@bayour.com>
+#       in function verifyChangesFile below.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Correction.
+###############################################################################
+
+sub getMD5HashForFile($) {
+    my ($file) = @_;
+    open(FILE, $file) or die "Can't open $file: $!\n";
+    binmode(FILE);
+    $digest = Digest::MD5->new->addfile(*FILE)->hexdigest, " $file\n";
+    close (FILE);
+    return $digest;
+}
+
 ###############################################################################
 # Name:		verifyChangesFile
 # Description:	Parses the .changes file.
@@ -1395,6 +1560,25 @@ sub handleDebOnlyFile($$) {
 #  2005-05-06 Daniel Leidert <daniel.leidert@wgdd.de>
 #       Add arg to handle signature verification in inputdir and distinput-dirs
 #       independetly.
+#  2007-10-08 Turbo Fredriksson <turbo@bayour.com>
+#       Stat only existing files.
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Correct indention (as Turbo explained).
+#  2007-10-08 Turbo Fredriksson <turbo@bayour.com>
+#       The changes file is only incomplete if the file is missing _and_
+#       a file in the destination do not already exist with the correct MD5
+#       hash.
+#  2007-10-08 Ola Lundqvist <opal@debian.org>
+#       Extraced the file check to two own functions.
+#       Changed the order to always check for installed files and reject if it
+#       exist with wrong size or md5sum.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
+#       Check incomplete time only once.
+#       Allow forcing an install of an upload (i.e., ignore existing files in
+#       destdir).
+#  2007-10-09 Ola Lundqvist <opal@debian.org>
+#       Changed force option to ignoredestcheck option.
 ###############################################################################
 
 sub verifyChangesFile($$) {
@@ -1402,45 +1586,70 @@ sub verifyChangesFile($$) {
     # Get changes file owner uid
     my $cfowner = (stat $cfile)[4];
     my $clastmod = (stat $cfile)[9];
-    # Verify changelog so it is complete, ok is default. If incomplete continue checking for errors.
+    my $old_changes = 0;
+    if ($clastmod < time() - $incompletetime) {
+	$old_changes = 1;
+    }
+    # Verify changelog so it is complete, ok is default.
+    # If incomplete continue checking for errors.
     my $ret = "ok";
-    foreach $file (keys %CFiles) {
-	if (! -f $file) {
-	    # Modification time of the changesfile.
-	    # 24*60*60 = 86400
-	    if ($clastmod < time() - 86400) {
-		pdebug(4, "$cfile is incomplete and is older than 24 hours.");
-		$CConf{ERROR} = "$CConf{ERROR}$cfile is incomplete and is older than 24 hours.\n";
-		return "reject";
+    foreach my $file (keys %CFiles) {
+	# Does it exist in the destdir?
+	my $inst_files = checkFileInDestDirs($file);
+	if (($inst_files ne "") && ($ignoredestcheck == 0)) {
+	    foreach my $inst_file (split / /, $inst_files) {
+		# Get data for the file to install.
+		my ($hash, $size, undef, undef) = parseFileStruct($CFiles{$file});
+		# Get size for the already installed one
+		my (undef, undef, undef, undef, undef, undef, undef, $ssize) = stat($inst_file);
+		if ($size != $ssize) {
+		    pdebug(4, "$cfile is rejected. $file already installed and size do not match.");
+		    $CConf{ERROR} = "$CConf{ERROR}$cfile is rejected. $file already installed and size do not match.\n";
+		    return "reject";
+		}
+		my $digest = getMD5HashForFile($inst_file);
+		if ($hash != $digest) {
+		    pdebug(4, "$cfile is rejected. $file already installed and MD5 sum do not match.");
+		    $CConf{ERROR} = "$CConf{ERROR}$cfile is rejected. $file already installed and MD5 sum do not match.\n";
+		    return "reject";
+		}
 	    }
-	    else {
+	}
+	elsif (! -f $file) {
+	    # Modification time of the changesfile.
+	    if ($old_changes) {
+		pdebug(4, "$cfile is incomplete and is older than $incompletetime seconds.");
+		$CConf{ERROR} = "$CConf{ERROR}$cfile is incomplete and is older than $incompletetime seconds.\n";
+		return "reject";
+	    } else {
+		pdebug(4, "$cfile is incomplete. Missing file: $file.");
 		$ret = "incomplete";
 	    }
-	}
-	my (undef, undef, undef, undef, $fowner, undef, undef, $size) = stat($file);
-	if ($fowner != $cfowner) {
-	    pdebug(4,"$file owner ($fowner) do not match $cfile owner ($cfowner).");
-	    $CConf{ERROR} = "$CConf{ERROR}$file owner do not match $cfile owner.\n";
-	    return "reject";
-	}
-	my ($cfhash, $cfsize) = split / /, $CFiles{$file};
-	if ($size < $cfsize) {
-	    # Modification time of the changesfile.
-	    # 24*60*60 = 86400
-	    pdebug(4, "$cfile is incomplete and is older than 24 hours as $file is not of full size.");
-	    if ($clastmod < time() - 86400) {
-		$CConf{ERROR} = "$CConf{ERROR}$cfile is incomplete and is older than 24 hours as $file is not of full size.\n";
+	} else {
+	    # File exist, now verify it.
+	    my (undef, undef, undef, undef, $fowner, undef, undef, $size) = stat($file);
+	    if ($fowner != $cfowner) {
+		pdebug(4,"$file owner ($fowner) do not match $cfile owner ($cfowner).");
+		$CConf{ERROR} = "$CConf{ERROR}$file owner do not match $cfile owner.\n";
 		return "reject";
 	    }
-	    else {
-		$ret = "incomplete";
+	    my ($cfhash, $cfsize) = split / /, $CFiles{$file};
+	    if ($size < $cfsize) {
+		# Modification time of the changesfile.
+		pdebug(4, "$cfile is incomplete and is older than $incompletetime seconds as $file is not of full size.");
+		if ($old_changes) {
+		    $CConf{ERROR} = "$CConf{ERROR}$cfile is incomplete and is older than $incompletetime seconds as $file is not of full size.\n";
+		    return "reject";
+		}
+		else {
+		    $ret = "incomplete";
+		}
+	    } elsif ($size > $cfsize) {
+		pdebug(4, "$cfile upload is broken, $file is bigger than expected.");
+		$CConf{ERROR} = "$CConf{ERROR}$cfile upload is broken, $file is bigger than expected.\n";
+		return "reject";
 	    }
-	} elsif ($size > $cfsize) {
-	    pdebug(4, "$cfile upload is broken, $file is bigger than expected.");
-	    $CConf{ERROR} = "$CConf{ERROR}$cfile upload is broken, $file is bigger than expected.\n";
-	    return "reject";
 	}
-	
     }
     # Verify signatures.
     if ($verify) {
@@ -1473,13 +1682,14 @@ sub verifyChangesFile($$) {
 #	make sure that CMeta is created before that.
 #  2005-04-xx Bob Proulx <bob@proulx.com>
 #	Split out parts of this function to parseChangesFile.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub handleChangesFile($) {
     my ($cfile) = @_;
 
-    my $file;
-    foreach $file (keys %CFiles) {
+    foreach my $file (keys %CFiles) {
 	handlePackageFile($file, $cfile);
     }
     installChangesFile($cfile);
@@ -1524,10 +1734,14 @@ sub uploaderIsChangesFileOwner($) {
 #	Modified reject cause.
 #  2005-05-06 Daniel Leidert <daniel.leidert@wgdd.de>
 #       Fixed move-command and check for .changes file.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
+#       Create reject log code.
+#  2007-10-09 Ola Lundqvist <opal@debian.org>
+#       Moved reject log code to REJECT dir and changed coding style some.
 ###############################################################################
 
 sub rejectChangesFile() {
-    my $file;
     if (! -d "REJECT") {
 	action(! mkpath ("REJECT", 0, 0775),
 	       "Making REJECT directory.",
@@ -1538,7 +1752,7 @@ sub rejectChangesFile() {
 		  "Move .changes to REJECT dir.",
 		  2);
     }
-    foreach $file (keys %CFiles) {
+    foreach my $file (keys %CFiles) {
 	if (-f $file) {
 	    cmdaction ("$movecmd $file REJECT/",
 		       "Move $file to REJECT dir.",
@@ -1548,6 +1762,18 @@ sub rejectChangesFile() {
 	    $CConf{ERROR} = "$CConf{ERROR}File $file can not be moved because it is already installed (or incomplete upload).\n";
 	}
     }
+    # Create a reject log
+    my $log = $CMeta{ChangesFile};
+    $log =~ s/\.changes/\.reject/;
+    $log = "REJECT/$log";
+    if (open(LOG, "> $log")) {
+	print LOG $CConf{ERROR};
+	close(LOG);
+    }
+    else {
+	pdebug(4, "Can't create reject log [$log].");
+	$CConf{ERROR} = "$CConf{ERROR}Can't create reject log [$log].\n";
+    }
 }
 
 ###############################################################################
@@ -1564,6 +1790,14 @@ sub rejectChangesFile() {
 #       Only remove source tarball file if not referenced by any other .changes file.
 #  2006-07-20 Ola Lundqvist <opal@debian.org>
 #       Changed indentation.
+#  2007-10-08 Turbo Fredriksson <turbo@bayour.com>
+#       Distmapping for files.
+#  2007-10-08 Turbo Fredriksson <turbo@bayour.com>
+#       The changes file is only incomplete if the file is missing _and_
+#       a file in the destination do not already exist with the correct MD5
+#       hash.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub handlePackageFile ($) {
@@ -1583,8 +1817,8 @@ sub handlePackageFile ($) {
 
     # OVERRIDES
 
-    my $distr;
-    foreach $distr (split /\s+/, $distrd) {
+    foreach my $distr (split /\s+/, $distrd) {
+	$distr = $distmapping{$distr} || $distr;
 	my $srcext = ".src" if ($arch =~ /^source$/);
 	parseOverrideFile($distr, $major, $srcext);
 	if (defined $Override{$distr, $major, "$pkgname$srcext"}) {
@@ -1614,7 +1848,7 @@ sub handlePackageFile ($) {
 		   "Making directory $installto",
 		   2);
 	}
-	else {
+	elsif (-f $file) {
 	    if ($arch =~ /^source$/) {
 		cmdaction("$rmcmd $installto/${pkgname}_*$ext",
 			  "Delete $installto/${pkgname}_*$ext",
@@ -1682,14 +1916,18 @@ sub handlePackageFile ($) {
 #	Wrote it.
 #  2003-03-14 Ola Lundqvist <opal@debian.org>
 #	Now CConf{ChangesFile} is changed when copying it.
+#  2007-10-08 Turbo Fredriksson <turbo@bayour.com>
+#       Distmapping for files.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub installChangesFile ($) {
     my ($cfile) = @_;
 
     my $distrd = $CConf{Distribution};
-    my $distr;
-    foreach $distr (split /\s+/, $distrd) {
+    foreach my $distr (split /\s+/, $distrd) {
+	$distr = $distmapping{$distr} || $distr;
 	my $todir = relativePath($cinstall, "$destdir/$distr");
 	if ($cinstall !~ /^\s*$/) {
 	    # Now remove or move away the .changes file (if $cinstall not empty).
@@ -1915,12 +2153,13 @@ sub parseFileName($) {
 #	Written with info from parseChanges.
 #  2003-02-12 Ola Lundqvist <opal@debian.org>
 #	Added ChangesFile to CConf hash.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub parseDebOnlyFile($$) {
     my ($kfile, $distr) = @_;
     my $state = "";
-    my $line;
     my $section;
     my $priority;
     my $size;
@@ -1930,7 +2169,7 @@ sub parseDebOnlyFile($$) {
     %CFiles = ();
     %CDesc = ();
     my @cmdres = readcommand("dpkg-deb -f $kfile");
-    foreach $line (@cmdres) {
+    foreach my $line (@cmdres) {
 	# The state to just put the line in the hash.
 	if ($line =~ /^\s*$/) {
 	    next;
@@ -2096,6 +2335,8 @@ sub parseChangesFile($) {
 # Changelog:
 #  2001-06-26 Ola Lundqvist <opal@debian.org>
 #	Written.
+#  2007-10-09 Turbo Fredriksson <turbo@bayour.com>
+#       Change in foreach to use my variable.
 ###############################################################################
 
 sub parseOverrideFile($$$) {
@@ -2109,9 +2350,8 @@ sub parseOverrideFile($$$) {
 	pdebug(5, "Load override file for $distr, $major");
 	my $odir = "$destdir/$distr/$major";
 	my @o = readfile("$odir/override$srcext");
-	my $tmp;
 	$Override{$distr,$major, $def} = 1;
-	foreach $tmp (@o) {
+	foreach my $tmp (@o) {
 	    my ($pkg, $prio, $section, $maint) = split(/\s+/, $tmp, 4);
 	    $pkg = "$pkg$srcext";
 	    $Override{$distr, $major, $pkg, Priority} = $prio
@@ -2275,6 +2515,15 @@ Default major section to use if undefined (default: main)
 
 Command to move files (currently not used at all).
 
+=item B<--incompletetime>
+
+The time to allow .changes file to be incomplete in seconds.
+Defaults to 24 hours.
+
+=item B<--mailfrom>
+
+Specify mail sender.
+
 =item B<--nosort>
 
 Do not sort packages.
@@ -2311,15 +2560,26 @@ On what level to quit the application, see debug level.
 
 Prints the version string.
 
+=item B<--ignoredestcheck>
+
+Force install of changes files even if some files already exists with wrong size or md5 hash.
+
 =head1 CONFIG FILE
 
 You can also place config files with the following names (in following order) /etc/debarchiver.conf, ~/.debarchiver.conf and input.conf (relative to input directory) that will be read before, the arguments to this program will be parsed. Here you can change the following variables:
 
+The configuration files are read as perl modules they should end with a true value. Therefore they should always end with 1;
+
 =item B<$bzip>
 
 If set to 0 no bzip2 files will be generated. If set to 1 bzip2 files will
 be generated.
 
+=item B<$incompletetime>
+
+Time to allow .changes files to be incomplete in seconds. Useful for slow uploads.
+Defaults to 24 hours.
+
 =item B<$cachedir>
 
 The cache directory for apt-ftparchive, used if --index is used.
@@ -2365,6 +2625,11 @@ The lockfile to use, default $lockfile.
 
 An array of strings that should be mailed to. If the string contains an email address that one is used. If it contains an incomplete email address, i.e. @hostname, the username owning the file is used @ the hostname specified. If no @ character is found in the string, it is considered as a field in the .changes file. Such a field can for example be Maintainer or Uploaders.
 
+=item B<$mailfrom>
+
+Specify the sender of emails.
+Default to none ("")
+
 =item B<$majordefault>
 
 Default major section (see --majordefault above).
@@ -2385,6 +2650,10 @@ The remove command (see --rmcmd above).
 
 Choose to enable (1) or disable (0) signature verification for packages uploaded into $inputdir (not %distinputdirs).
 
+=item B<$ignoredestcheck>
+
+Force install of changes files even if some files already exists with wrong size or md5 hash. Default to 0.
+
 =item B<$verifysignaturesdistinput>
 
 Choose to enable (1) or disable (2) signature verification for packages uploaded into %distinputdirs. This works independently from $verifysignatures.
@@ -2403,6 +2672,15 @@ You should use either B<--autoscanpackages> and B<--autoscansources> or B<--inde
 
 The default action (and the recommended) is B<--scandetect> that probe for installed software and use the best choice depending on what software you have installed (choose between --index and --autoscan right now).
 
+=head1 REJECT
+
+Changes files are rejected in the following conditions:
+ * If a file that is about to be installed, already exist in the archive and is not identical to the one that is about to be installed.
+ * Changes file is incomplete and has been there for $incompletetime time.
+ * A file that is part of the Changes file is not yet big enough and the changes file has been there for $incompletetime time.
+ * A file that is part of the Changes file is bigger than specified.
+ * If verify signatures is enabled and signature do not match.
+
 =head1 FILES
 
 B</etc/debarchiver.conf>
